local filter = ...
checkArg(1, filter, 'string', 'nil')

local invoke = component.invoke

local src = computer.getBootAddress()

local candidates = {}
for fs in component.list('filesystem') do
	if
		fs ~= src
		and fs ~= computer.tmpAddress()
		and not invoke(fs, 'isReadOnly')
		and (not filter or fs:sub(1, #filter) == filter)
	then
		table.insert(candidates, fs)
	end
end

assert(#candidates > 0, 'no filesystem to install to')
if #candidates > 1 then
	for i, addr in ipairs(candidates) do candidates[i] = '- ' .. addr end
	candidates[0] = 'multiple candidates to install to:'
	error(table.concat(candidates, '\n', 0, #candidates))
end

local dst = candidates[1]

local lastYield = 0
local function yield()
	if computer.uptime() - lastYield > 1 then
		os.sleep(0)
		lastYield = computer.uptime()
	end
end

local exclude = {
	['/etc/init.lua'] = true,
}

local function copy(path)
	yield()
	if exclude[path] and invoke(dst, 'exists', path) then
		print('skipping ' .. path)
		return true
	end
	local ls = invoke(src, 'list', path)
	if ls then
		table.sort(ls) --for pretty output
		if path:sub(-1) ~= '/' then path = path .. '/' end
		local ok, err = invoke(dst, 'makeDirectory', path)
		assert(ok or invoke(dst, 'isDirectory', path), err or 'can\'t create ' .. path)
		for i, child in ipairs(ls) do copy(path .. child) end
		return true
	end

	--filesystem.list() returns no values if the path doesn't exist or is a file. assume the file exists

	local srcfile, err = io.open(path, 'rb')
	if not srcfile then
		error(string.format('reading %s on %s: %s', path, src:sub(1, 8), err))
	end
	local dstfile, err = invoke(dst, 'open', path, 'w')
	if not dstfile then
		srcfile:close()
		error(string.format('writing %s on %s: %s', path, dst:sub(1, 8), err))
	end

	local size, suffix = assert(invoke(src, 'size', path)), 'B'
	local suffix
	if size >= 1024 * 1024 then
		size, suffix = size // (1024 * 1024), 'MB'
	elseif size >= 1024 then
		size, suffix = size // 1024, 'kB'
	end
	print(string.format('copying %s (%d%s)', path, size, suffix))

	local chunk, err
	repeat
		yield()
		local sz, free = size, computer.freeMemory()
		if sz > free then sz = free end
		if sz < free / 4 then size = free // 4 end
		chunk, err = srcfile:read(sz)

		if chunk then chunk, err = invoke(dst, 'write', dstfile, chunk) end
	until not chunk

	invoke(dst, 'close', dstfile)
	srcfile:close()
	if err then error(string.format('copying %s: %s', path, err)) end

	return true
end

copy('/')

-- vi: set ts=2:
