--provide a tty on each available screen, running the given program
local progArgs = table.pack(...)
checkArg(1, progArgs[1], 'string') --program path


--identify the highest-tier gpu available
local gpu, vmem --use virtual memory as a heuristic for tier
for addr in component.list('gpu', true) do
	local vmemNew = component.invoke(addr, 'totalMemory')
	if vmemNew > (vmem or -1) then
		gpu, vmem = addr, vmemNew
	end
end
gpu = assert(gpu, 'please install a GPU or APU to continue')

--track most recently assigned values to skip redundant bind/setBackground/setForeground calls
local curScreen, curBg, curFg
local function bind(screen)
	if screen ~= curScreen then
		component.invoke(gpu, 'bind', screen)
		curScreen = screen
	end
end
local function setBackground(color)
	if color ~= curBg then
		component.invoke(gpu, 'setBackground', color)
		curBg = color
	end
end
local function setForeground(color)
	if color ~= curFg then
		component.invoke(gpu, 'setForeground', color)
		curFg = color
	end
end
local function draw(cx, text)
	bind(cx.screen)
	setBackground(cx.bg)
	setForeground(cx.fg)
	return component.invoke(gpu, 'set', cx.x, cx.y, text)
end

local iomt = {}
iomt.__index = iomt

local function isAttachedKeyboard(cx, addr)
	for i, kb in ipairs(cx.keyboards) do
		if kb == addr then
			return i
		end
	end
end

--tail-recursively pullSignal to construct the line of input read from a terminal
local function readSignal(cx, input, sig, ...)
	if sig == 'key_down' then
		local kb, ch, code = ...
		if isAttachedKeyboard(cx, kb) then
			if code == 28 then --enter
				cx.x, cx.y = 1, cx.y + 1
				return input .. '\n'
			elseif code == 14 then --backspace
				if input ~= '' then
					cx.x = cx.x - 1
					draw(cx, ' ')
					input = input:sub(1, -2)
				end
			elseif ch ~= 0 then
				local char = string.char(ch)
				draw(cx, char)
				cx.x = cx.x + 1
				input = input .. char
			end
		end
	elseif sig == 'component_added' then
		local addr, ty = ...
		if ty == 'keyboard' then
			cx.keyboards = component.invoke(cx.screen, 'getKeyboards')
		end
	elseif sig == 'component_removed' then
		local addr, ty = ...
		if addr == cx.screen then
			return nil, 'disconnected'
		end
		local idx = ty == 'keyboard' and isAttachedKeyboard(cx, addr)
		if idx then
			table.remove(cx.keyboards, idx)
		end
	end

	return readSignal(cx, input, computer.pullSignal())
end

-- file handle methods
function iomt:read(n)
	--technically this should return at maximum `n` bytes, but that would involve a redundant second buffer
	--we happen to know the implementation works fine returning more than `n`

	return readSignal(self, '', computer.pullSignal())
end

function iomt:write(data)
	bind(self.screen) setBackground(self.bg) setForeground(self.fg)
	local idx = 1
	::line::
	local lf = data:find('\n', idx, true)
	component.invoke(gpu, 'set', self.x, self.y, data:sub(idx, lf and lf - 1))

	if lf then
		self.x, self.y = 1, self.y + 1
		idx = lf + 1
		goto line
	end

	self.x = self.x + #data - idx + 1
	return true
end

--no seek, noop close
function iomt:close() return true end

-- tty methods
function iomt:mark()
	return { bg = self.bg, fg = self.fg }
end
function iomt:reset(mark)
	self.bg, self.fg = mark.bg, mark.fg
end
function iomt:setBg(color) self.bg = color end
function iomt:setFg(color) self.fg = color end


local function createTerm(screen, prog, ...)
	checkArg(1, screen, 'string')
	checkArg(2, prog, 'string')
	local fh = setmetatable({ screen = screen, keyboards = component.invoke(screen, 'getKeyboards'), x = 1, y = 1, fg = 0xffffff, bg = 0x000000 }, iomt)
	local stdio = assert(kitn.wrapFile('r+', fh))
	stdio.tty = fh

	bind(screen)
	setForeground(0xffffff)
	setBackground(0x000000)
	local w, h = component.invoke(gpu, 'maxResolution')
	component.invoke(gpu, 'setResolution', w, h)
	component.invoke(gpu, 'fill', 1, 1, w, h, ' ')

	local process = assert(kitn.createProcess(prog, stdio, ...)) --TODO set io via createProcess args
	return process, stdio
end



--for each existing screen, spawn prog
for screen in component.list('screen') do
	createTerm(screen, table.unpack(progArgs, 1, progArgs.n))
end

--when new screens are added, spawn prog on those too
local function tick(sig, addr, ty)
	if sig == 'component_added' and ty == 'screen' then
		createTerm(addr, table.unpack(progArgs, 1, progArgs.n))
	end

	return tick(computer.pullSignal())
end

return tick(computer.pullSignal())

-- vi: set ts=2:
