local function kreadfile(path, fs)
	fs = fs or bootfs
	local text, chunk, fd, err = '', nil, invoke(fs, 'open', path)
	if not fd then
		local msg = format('can\'t open %s on %s', path, fs:sub(1, 3))
		if not rawequal(path, err) then msg = msg .. ': ' .. err end
		return nil, msg
	end

	repeat
		chunk, err = invoke(fs, 'read', fd, big)
		text = text .. (chunk or '')
	until not chunk
	invoke(fs, 'close', fd)
	if err then return nil, format('reading %s: %s', path, err) end
	return text
end


--[[table of process information. key is the main thread's coroutine, value is a table with:
- array part: coroutines for child threads and processes
- `parent`: coroutine of parent process
- `env`: process environment
- `path`: filesystem path the process was loaded from (for diagnostics)
]]
local processes = {}

--run queue. key is coroutine (process or thread), value is pack()ed args to resume it with.
--`runqueue[co]` is temporarily set to `false` while the thread is running.
local runqueue = {}

--tasks waiting for a timeout. key is coroutine (process or thread), value is absolute deadline (from uptime()).
local schedDeadline = {}

--tasks waiting for a signal. key is coroutine (process or thread), value is `true` for all signals (may support specific signals in the future).
local schedSignal = {}


local fillEnv --function(proc:thread, env:table):() -- given a process and an empty environment, fill it in before the process runs

local function createProcess(path, parent)
	local code, err = kreadfile(path)
	if not code then return nil, err end
	local env = {}

	--using kload here is okay because we always pass a non-nil environment
	local f, err = kload(code, '@' .. path, 'bt', env)
	if not f then return nil, err end

	local co = co_create(f)
	fillEnv(co, env)

	local info = { env = env, parent = parent, path = path }
	processes[co] = info

	return co, info
end

-- vi: set ts=2:
