--[[environments for new processes
given <https://ocdoc.cil.li/tutorial:custom_oses#what_s_available>:
- replace load to default to the current process's environment, not the global environment
- provide [loadfile, dofile] using the boot filesystem and the current process's environment
- replace computer.pullSignal and provide os.sleep to yield from the current coroutine, not the entire computer
- replace coroutine.yield to differentiate its yields from os yields (pullSignal/sleep)
- replace coroutine.resume to pass os yields through, suspending the os thread and not just the current coroutine
- provide kitn namespace for os functions
]]


envBase._OSVERSION = 'kitn 0.2.1-beta.5' --prefer [openloader, openos] "$name $ver" over plan9k "$name/$ver"
envBase.kitn = {}
envBase._G, envBase.load = nil

do
--[[propagating os yields
	when a process calls coroutine.yield(), we prepend `false` to the yielded values. os yields start with a truthy value.
	in coroutine.resume(), if the coroutine yielded (ie isn't dead) and the first value is truthy (ie an os yield),
	we also yield from the current coroutine, unwinding the stack of running coroutines back to the os thread.
	when the current coroutine is resumed, we pass the values back into the inner coroutine and repeat.
]]
	envBase.coroutine.yield = function(...) return kyield(false, ...) end

	local forwardYield
	envBase.coroutine.resume = function(co, ...)
		checkArg(1, co, 'thread')

		return forwardYield(co, kresume(co, ...))
	end

	function forwardYield(co, ok, ...)
		if co_status(co) ~= 'suspended' then --not an os yield if the coroutine didn't yield (because it returned/errored)
			return ok, ...
		elseif (...) then --first arg is truthy for os yield, false for coroutine.yield()
			--yield to the scheduler, pass the results (ie signal) back to `co`, then repeat for `co`'s next yield
			return forwardYield(co, kresume(co, kyield(...))) 
		else --`co` called coroutine.yield(), so return its values to the waiting coroutine.resume()
			return ok, select(2, ...)
		end
	end
end

envBase.os.sleep = function(n)
	checkArg(1, n, 'number')
	return kyield('deadline', n)
end
envBase.computer.pullSignal = function(timeout)
	checkArg(1, timeout, 'number', 'nil')
	return kyield('signal', timeout)
end


function fillEnv(proc, uenv)
	mirror(envBase, uenv)
	uenv._G = uenv

	--define these in local variables so the process deleting `load` doesn't break its `loadfile`
	local function uload(chunk, name, mode, env)
		checkArg(1, chunk, 'function', 'string')
		checkArg(2, name, 'string', 'nil')
		checkArg(3, mode, 'string', 'nil')
		checkArg(4, env, 'table', 'nil')
		return kload(chunk, name, mode, env or uenv)
	end
	local function uloadfile(path, mode, env)
		checkArg(1, path, 'string')
		checkArg(2, mode, 'string', 'nil')
		checkArg(3, env, 'table', 'nil')

		local code, err = kreadfile(path)
		if not code then return nil, err end

		code, err = uload(code, '@' .. path, mode, env)
		--`return code, err` here would return 2 values on success, unlike vanilla loadfile
		if not code then return nil, err end
		return code
	end
	local function udofile(path, ...)
		checkArg(1, path, 'string')
		local fn = assert(uloadfile(path))
		return fn(...)
	end

	uenv.load, uenv.loadfile, uenv.dofile = uload, uloadfile, udofile

	local stdin, stdout = openNull('r'), openNull('w')
	local uio; uio = {
		stdin = stdin, stdout = stdout, stderr = stdout,
		close = function(file) return (file or uio.stdout):close() end,
		flush = function() return uio.stdout:flush() end,
		input = function(file)
			local t = checkArgEx(1, file, 'string', 'table', 'nil')
			if file then
				uio.stdin = (t == 'string' and assert(uio.open(file, 'r')) or file)
			end
			return uio.stdin
		end,
		lines = function(filename, ...)
			checkArg(1, filename, 'string', 'nil')
			error('TODO io.lines') --iff we open a file, we need to close it
		end,
		open = function(filename, mode)
			checkArg(1, filename, 'string')
			checkArg(2, filename, 'string', 'nil')
			return openPath(mode, bootfs, filename)
		end,
		output = function(file)
			local t = checkArgEx(1, file, 'string', 'table', 'nil')
			if file then
				uio.stdout = (t == 'string' and assert(uio.open(file, 'w')) or file)
			end
			return uio.stdout
		end,
		--popen is unimplemented because we don't have a notion of a shell
		read = function(...) return uio.stdin:read(...) end,
		tmpfile = function()
			local deadline = uptime() + 1 --give up after 1s instead of potentially infinitely looping
			local filename
			repeat
				filename = format('tmp.%08x', random(0, 0x7fffffff))
				if invoke(tmpfs, 'exists', filename) then filename = nil end
			until filename or uptime() >= deadline
			assert(filename, 'failed to allocate tmpfile')
			return openPath('w+', tmpfs, filename) --TODO close at exit
		end,
		type = function(obj)
			local file, opened = isFile(obj)
			if opened then
				return 'file'
			elseif file then
				return 'closed file'
			else
				return nil
			end
		end,
		write = function(...) return uio.stdout:write(...) end,
	}
	uenv.io = uio

	function uenv.print(...)
		local parts = pack(...)
		for i = 1, parts.n do
			parts[i] = tostring(parts[i])
		end
		assert(uenv.io.write(concat(parts, '\t', 1, parts.n), '\n'))
	end


	--env will automatically mirror envBase.kitn here
	local ukitn = uenv.kitn

	function ukitn.createThread(fn, ...)
		checkArg(1, fn, 'function')
		local co = co_create(fn)
		local info = processes[proc]
		info[#info+1] = co --add it as a thread of the current process
		runqueue[co] = pack(...)
		return co
	end

	function ukitn.createProcess(path, ...)
		checkArg(1, path, 'string')
		local co, info = createProcess(path, proc)
		if not co then return nil, info end
		runqueue[co] = pack(...)
		return co
	end

	function ukitn.running()
		--TODO return name, child processes, child threads, ...
		--TODO return information for the current thread within the process too
		return proc, nil
	end

	--TODO map all filesystems into a single tree, then remove this method
	function ukitn.openOn(address, path, mode)
		checkArg(1, address, 'string')
		checkArg(2, path, 'string')
		checkArg(3, mode, 'string', 'nil')
		local t, err = component_type(address)
		if t ~= 'filesystem' then
			return nil, err or 'component is not a filesystem'
		end
		return openPath(mode, address, path)
	end

	local function wrap_read(fh, ...) return fh.stream:read(...) end
	local function wrap_write(fh, ...) return fh.stream:write(...) end
	local function wrap_seek(fh, ...) return fh.stream:seek(...) end
	local function wrap_close(fh, ...) return fh.stream:close(...) end

	--TODO expose this as a require()able library like <https://ocdoc.cil.li/api:buffer>
	function ukitn.wrapFile(mode, stream)
		checkArg(1, mode, 'string')
		checkArg(2, stream, 'table')
		--in the kernel, we can take a shortcut and add _file_* attributes directly to the caller's (our own) fh. we don't
		--want to do that to arbitrary userspace values though, so we add another layer of indirection.
		return wrapFile(mode, { stream = stream, read = wrap_read, write = wrap_write, seek = wrap_seek, close = wrap_close })
	end
end

-- vi: set ts=2:
