local mt = {}
mt.__index = mt

--wrap the current process's stdio streams into a tty that supports styled output.
local function wrap(io_)
	checkArg(1, io_, 'table', 'nil')
	io_ = io_ or io

	--cache the streams here so that `local tty = wrap(io); io.output('/some/file'); tty.prompt()` works as expected
	local stdin, stdout, stderr = io_.stdin, io_.stdout, io_.stderr
	if stdout.tty then stdout:setvbuf('no') end
	if stderr.tty then stderr:setvbuf('no') end

	return setmetatable({ stdin = stdin, stdout = stdout, stderr = stderr }, mt)
end

local function writeStyled(tty, stream, arg)
	if type(arg) == 'table' then
		local token = tty:mark()
		if arg.bg then tty:setBg(arg.bg) end
		if arg.fg then tty:setFg(arg.fg) end
		for i = 1, arg.n or #arg do writeStyled(tty, stream, arg[i]) end
		stream:flush() --ensure
		tty:reset(token)
	else
		stream:write(arg)
	end
end

local function writeUnstyled(stream, arg)
	if type(arg) == 'table' then
		for i = 1, arg.n or #arg do
			writeUnstyled(stream, arg[i])
		end
	else
		stream:write(arg)
	end
end

--[[write strings to stdout with optional styling.

tables are assumed to be styled groups, ie sequences of strings/styled groups with optional extra properties:
	.fg = 0xrrggbb sets foreground color for all contained text, unless overridden by a styled group inside
	.bg = 0xrrggbb sets background color for all contained text, unless overridden by a styled group inside
these properties are applied while rendering the styled group, then reverted to the prior values before rendering
any following values in this or future write() calls.

if stdout is not a tty, all styling is ignored.
]]
function mt:write(...)
	local args = table.pack(...)
	local caps = self.stdout.tty
	if caps then
		writeStyled(caps, self.stdout, args)
	else
		writeUnstyled(self.stdout, args)
	end
	return self
end

--write strings to stderr with optional styling.
--see `mt:write` for how styling is interpreted.
function mt:error(...)
	--TODO use wrapped stderr
	return self:write({ fg = 0xff0000, ... })
end

--prompt the user for a line of input.
--if stdout isn't a tty, equivalent to `io.stdin:read('l')`.
function mt:prompt(...)
	if self.stdin.tty and self.stdout.tty then
		self:write(...)
		self.stdout:flush()
	end
	return self.stdin:read('l')
end

return { wrap = wrap }

-- vi: set ts=2:
